#include <errno.h>

#include "emu32.h"
#include "emu-hcall.c"

/* --------------------------------------------------------------------- */

#define HCALL_HANDLED  0
#define HCALL_FORWARD -1

typedef int32_t (*xen_hypercall)(uint32_t *args);
static int32_t multicall(uint32_t *args);

/* --------------------------------------------------------------------- */

static int32_t update_va_mapping(uint32_t *args)
{
    uint32_t va    = args[0];
    uint64_t val   = args[1] | ((uint64_t)args[2] << 32 );
    uint32_t flags = args[3];

    if (is_pae()) {
	uint64_t *pte = find_pte_pae(va);

	if ((*pte & ~_PAGE_RW) != val || (val & _PAGE_USER)) {
	    *pte = val;
#if 0
	} else {
	    printk("%s: ignoring mkreadonly (va 0x%" PRIx32 ")\n",
		   __FUNCTION__, va);
#endif
	}
    } else {
	uint32_t *pte = find_pte_32(va);

	if ((*pte & ~_PAGE_RW) != val || (val & _PAGE_USER)) {
	    *pte = val;
#if 0
	} else {
	    printk("%s: ignoring mkreadonly (va 0x%" PRIx32 ")\n",
		   __FUNCTION__, va);
#endif
	}
    }

    switch (flags & UVMF_FLUSHTYPE_MASK) {
    case UVMF_NONE:
	break;
    case UVMF_TLB_FLUSH:
	flush_tlb();
	break;
    case UVMF_INVLPG:
	flush_tlb_addr(va);
	break;
    }
    return 0;
}

static int32_t mmu_update(uint32_t *args)
{
    uint64_t *reqs = (void*)args[0];
    uint32_t count = args[1];
//    uint32_t *done = (void*)args[2];
    uint32_t dom   = args[3];
    int i;

    if (dom != DOMID_SELF) {
	printk("%s: foreigndom not supported\n", __FUNCTION__);
	return -ENOSYS;
    }

    for (i = 0; i < count; i++) {
	switch (reqs[0] & 3) {
	case MMU_NORMAL_PT_UPDATE:
	    if (is_pae()) {
		uint64_t *pte = map_pae(reqs[0]);
		*pte = reqs[1];
	    } else {
		uint32_t *pte = map_32(reqs[0]);
		*pte = reqs[1];
	    }
	    break;
	case MMU_MACHPHYS_UPDATE:
	    printk("%s: machphys: %" PRIx64 " %" PRIx64 "\n",
		   __FUNCTION__, reqs[0], reqs[1]);
	    panic("FIXME: MMU_MACHPHYS_UPDATE not implemented yet");
	    break;
	default:
	    return -ENOSYS;
	}
	reqs += 2;
    }
    return 0;
}

static int32_t mmuext_op(uint32_t *args)
{
    struct mmuext_op *uops = (void*)args[0];
    uint32_t count = args[1];
//    uint32_t *done = (void*)args[2];
    uint32_t dom   = args[3];
    int i;

    if (dom != DOMID_SELF) {
	printk("%s: foreigndom not supported\n", __FUNCTION__);
	return -ENOSYS;
    }

    for (i = 0; i < count; i++, uops++) {
	switch (uops->cmd) {
	case MMUEXT_PIN_L1_TABLE:
	case MMUEXT_PIN_L2_TABLE:
	case MMUEXT_PIN_L3_TABLE:
	case MMUEXT_PIN_L4_TABLE:
	case MMUEXT_UNPIN_TABLE:
	    /* ignore */
	    break;
	case MMUEXT_SET_LDT:
	    printk("%s: FIXME: ignore SET_LDT\n", __FUNCTION__);
	    break;
	case MMUEXT_INVLPG_LOCAL:
	    flush_tlb_addr(uops->arg1.linear_addr);
	    break;
	case MMUEXT_TLB_FLUSH_LOCAL:
	case MMUEXT_TLB_FLUSH_ALL:    /* no SMP yet */
	case MMUEXT_TLB_FLUSH_MULTI:  /* no SMP yet */
	    flush_tlb();
	    break;
	case MMUEXT_NEW_BASEPTR:
	    if (is_pae())
		update_emu_mappings_pae(uops->arg1.mfn);
	    else
		update_emu_mappings_32(uops->arg1.mfn);
	    write_cr3(frame_to_addr(uops->arg1.mfn));
	    break;
	default:
	    printk("%s: FIXME: unknown %d\n", __FUNCTION__, uops->cmd);
	    return -ENOSYS;
	}
    }
    return 0;
}

/* --------------------------------------------------------------------- */

static xen_hypercall hcalls[XEN_HCALL_MAX] = {
    [ __HYPERVISOR_update_va_mapping ] = update_va_mapping,
    [ __HYPERVISOR_mmu_update ]        = mmu_update,
    [ __HYPERVISOR_mmuext_op ]         = mmuext_op,
    [ __HYPERVISOR_update_descriptor ] = update_descriptor,
    [ __HYPERVISOR_stack_switch ]      = stack_switch,
    [ __HYPERVISOR_multicall ]         = multicall,
    [ __HYPERVISOR_physdev_op ]        = physdev_op,
    [ __HYPERVISOR_physdev_op_compat ] = physdev_op,
    [ __HYPERVISOR_fpu_taskswitch ]    = fpu_taskswitch,
};

static int32_t multicall(uint32_t *args)
{
    struct multicall_entry *calls = (void*)args[0];
    uint32_t i, count = args[1];
    uint32_t margs[6];

    for (i = 0; i < count; i++) {
	if (!hcalls[calls[i].op])
	    panic("unknown hypercall in multicall list");
	emu_mcalls[calls[i].op]++;
	margs[0] = calls[i].args[0];
	margs[1] = calls[i].args[1];
	margs[2] = calls[i].args[2];
	margs[3] = calls[i].args[3];
	margs[4] = calls[i].args[4];
	margs[5] = calls[i].args[5];
	calls[i].result = hcalls[calls[i].op](margs);
    }
    return 0;
}

asmlinkage void do_hypercall(struct regs_32 regs)
{
    uint32_t args[6];
    uint32_t retval = -ENOSYS;

    if (regs.eax >= XEN_HCALL_MAX) {
	/* invalid hypercall number */
	goto handled;
    }
    if (!hcalls[regs.eax]) {
	/* no hypercall handler */
	goto forward;
    }

    /* do call */
    emu_hcalls[regs.eax]++;
    args[0] = regs.ebx;
    args[1] = regs.ecx;
    args[2] = regs.edx;
    args[3] = regs.esi;
    args[4] = regs.edi;
    args[5] = regs.ebp;
    retval = hcalls[regs.eax](args);
    if (-ENOSYS == retval)
	goto forward;

handled:
    regs.error = HCALL_HANDLED;
    regs.eax = retval;
    return;

forward:
    regs.error = HCALL_FORWARD;
    return;
}
